#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <zlib.h>

#include "data.h"
#include "toc.h"

int getsize(FILE *f)
{
	int size;

	fseek(f, 0, SEEK_END);
	size = ftell(f);

	fseek(f, 0, SEEK_SET);
	return size;
}

void ErrorExit(char *fmt, ...)
{
	va_list list;
	char msg[256];

	va_start(list, fmt);
	vsprintf(msg, fmt, list);
	va_end(list);

	printf(msg);
	exit(-1);
}

z_stream z;

int deflateCompress(void *inbuf, int insize, void *outbuf, int outsize, int level)
{
	int res;

	z.zalloc = Z_NULL;
	z.zfree  = Z_NULL;
	z.opaque = Z_NULL;

	if (deflateInit2(&z, level , Z_DEFLATED, -15, 8, Z_DEFAULT_STRATEGY) != Z_OK)
		return -1;

	z.next_out  = (Bytef *)outbuf;
	z.avail_out = outsize;
	z.next_in   = (Bytef *)inbuf;
	z.avail_in  = insize;

	if (deflate(&z, Z_FINISH) != Z_STREAM_END)
	{
		return -1;
	}

	res = outsize - z.avail_out;

	if (deflateEnd(&z) != Z_OK)
		return -1;

	return res;
}

int inflateDecompress(void *inbuf, int insize, void *outbuf, int outsize)
{
	z.zalloc = Z_NULL;
	z.zfree  = Z_NULL;
	z.opaque = Z_NULL;

	if (inflateInit2(&z, -15) != Z_OK)
		return -1;

	z.next_in   = (Bytef *)inbuf;
	z.avail_in  = insize;
	z.next_out  = (Bytef *)outbuf;
	z.avail_out = outsize;

	if (inflate(&z, Z_FINISH) != Z_STREAM_END)
      return -1;

	if (inflateEnd(&z) != Z_OK)
		return -1;

	return 0;
}

#ifndef WIN32
typedef struct  __attribute__((packed))
#else
#pragma pack(1)
typedef struct
#endif
{
	unsigned int signature;
	unsigned int version;
	unsigned int fields_table_offs;
	unsigned int values_table_offs;
	int nitems;
} SFOHeader;
#ifdef WIN32
#pragma pack()
#endif

#ifndef WIN32
typedef struct  __attribute__((packed))
#else
#pragma pack(1)
typedef struct
#endif
{
	unsigned short field_offs;
	unsigned char  unk;
	unsigned char  type; // 0x2 -> string, 0x4 -> number
	unsigned int length;
	unsigned int size;
	unsigned short val_offs;
	unsigned short unk4;
} SFODir;
#ifdef WIN32
#pragma pack()
#endif

char buffer[1*1048576];
char buffer2[0x9300];

int sfo = 0,pic0 = 0, pic1 = 0, icon0 = 0, icon1 = 0, snd = 0,prx = 0, toc = 0,splash=0;
int sfo_size, pic0_size, pic1_size, icon0_size, icon1_size, snd_size, toc_size, prx_size,splash_size;
int start_dat = 0;
unsigned int header[0x28/4];
unsigned int dummy[6];

#ifndef WIN32
typedef struct __attribute__((packed))
#else
#pragma pack(1)
typedef struct
#endif
{
	unsigned int offset;
	unsigned int length;
	unsigned int dummy[6];
} IsoIndex;
#ifdef WIN32
#pragma pack()
#endif
//SFOf[^---------------------------------------------------------------
char sfo_header[8]={0x00,0x50,0x53,0x46,0x01,0x01,0x00,0x00};//\0PSF\1\1\0\0

int pm_n=6;
char pm_title[6][0x10]={
	"BOOTABLE",
	"CATEGORY",
	"DISC_ID",
	"PARENTAL_LEVEL",
	"REGION",
	"TITLE"
};
char pm_type[6]={1,0,0,1,1,0};
int pm_len[6]={0x4,0x4,0x10,0x4,0x4,0x80};
int pm_b[6]={1,0,0,0,0,0};
char pm_s[6][0x80]={
	"",
	"ME",
	"",
	"",
	"",
	""
};
//-----------------------------------------------------------------------
int CreateSFO(){
	FILE *out = fopen("PARAM.SFO","wb");
	if(!out){
		fclose(out);
		return 0;		
	}
	
	//PARAM.SFO쐬
    int sa=0;
    int po=0;
	int i=0;
    fwrite(sfo_header,1,sizeof(sfo_header),out);
        
    int str_ad=sizeof(sfo_header)+4+4+4+16*pm_n;
    fwrite(&str_ad,1,4,out);
    int byte_ad=0;
    for(i=0;i<pm_n;i++){
    	byte_ad+=strlen(pm_title[i])+1;
    }
        
    while((str_ad+byte_ad)%0x4!=0){
        byte_ad++;
    	sa++;
    }
    po=(str_ad+byte_ad);
    fwrite(&po,1,4,out);
    fwrite(&pm_n,1,4,out);
    
    short str_offset=0;
    int byte_offset=0;
    
    for(i=0;i<pm_n;i++){
        fwrite(&str_offset,1,2,out);
        po=4;
        fwrite(&po,1,1,out);
        if(pm_type[i]){
        	po=4;
        	fwrite(&po,1,1,out);
        }else{
        	po=2;
        	fwrite(&po,1,1,out);
        }
        fwrite(&pm_len[i],1,4,out);
        fwrite(&pm_len[i],1,4,out);
        fwrite(&byte_offset,1,4,out);
        
        str_offset+=strlen(pm_title[i])+1;
        byte_offset+=pm_len[i];
    }
    for(i=0;i<pm_n;i++){
    	fwrite(pm_title[i],1,strlen(pm_title[i])+1,out);
    }
    for(i=0;i<sa;i++){
    	fputc(0, out);
    }
    for(i=0;i<pm_n;i++){
        if(pm_type[i]){
        	fwrite(&pm_b[i],1,pm_len[i],out);
        }else{
        	fwrite(pm_s[i],1,pm_len[i],out);
        }
    }
	
	fclose(out);
	return 1;
}
void convert(char *input, char *output, char *title, char *code, int complevel)
{
	FILE *in, *out, *t;
	int i, offset, isosize, isorealsize, x;
	int index_offset, p1_offset, p2_offset, end_offset;
	IsoIndex *indexes;

	void* tocptr;
	in = fopen (input, "rb");
	if (!in){
		ErrorExit("%sJ܂B\n", input);
	}

	isosize = getsize(in);
	isorealsize = isosize;

	if ((isosize % 0x9300) != 0)
	{
		isosize = isosize + (0x9300 - (isosize%0x9300));
	}

	//printf("isosize, isorealsize %08X  %08X\n", isosize, isorealsize);

	out = fopen(output, "wb");
    if (!out){
        ErrorExit("%s쐬邱Ƃł܂B\n", output);
    }

	printf("wb_[ł܂...\n");

	if(CreateSFO()==1){
		t = fopen("PARAM.SFO", "rb");
		sfo_size = getsize(t);
		sfo = 1;
		fclose(t);
	}
	t = fopen("../ICON0.PNG", "rb");
	if (t)
	{
		icon0_size = getsize(t);
		icon0 = 1;
		fclose(t);
	}
	else
	{
		icon0_size = 0;
	}

	t = fopen("../ICON1.PMF", "rb");
	if (t)
	{
		icon1_size = getsize(t);
		icon1 = 1;
		fclose(t);
	}
	else
	{
		icon1_size = 0;
	}

	t = fopen("../PIC0.PNG", "rb");
	if (t)
	{
		pic0_size = getsize(t);
		pic0 = 1;
		fclose(t);
	}
	else
	{
		pic0_size = 0; //base_header[6] - base_header[5];
	}

	t = fopen("../PIC1.PNG", "rb");
	if (t)
	{
		pic1_size = getsize(t);
		pic1 = 1;
		fclose(t);
	}
	else
	{
		pic1_size = 0; // base_header[7] - base_header[6];
	}

	t = fopen("../SND0.AT3", "rb");
	if (t)
	{
		snd_size = getsize(t);
		snd = 1;
		fclose(t);
	}
	else
	{
		snd = 0;
	}

	t = fopen("DATA.PSP", "rb");
	if (t)
	{
		prx_size = getsize(t);
		prx = 1;
		fclose(t);
	}else{
		prx_size = 0;
	}

	if ((tocptr = create_toc(input, &toc_size)) != NULL)
	{
		toc = 2;
	}
	else
	{
		toc = 0;
	}

	int curoffs = 0x28;

	header[0] = 0x50425000;
	header[1] = 0x10000;

	header[2] = curoffs;

	curoffs += sfo_size;
	header[3] = curoffs;

	curoffs += icon0_size;
	header[4] = curoffs;

	curoffs += icon1_size;
	header[5] = curoffs;

	curoffs += pic0_size;
	header[6] = curoffs;

	curoffs += pic1_size;
	header[7] = curoffs;

	curoffs += snd_size;
	header[8] = curoffs;

	x = header[8] + prx_size;

	if ((x % 0x10000) != 0)
	{
		x = x + (0x10000 - (x % 0x10000));
	}

	header[9] = x;

	fwrite(header, 1, 0x28, out);


	printf("param.sfoł܂...\n");
	if (sfo){
		t = fopen("PARAM.SFO", "rb");
    	fread(buffer, 1, sfo_size, t);
    	fwrite(buffer, 1, sfo_size, out);
    	fclose(t);
		remove("PARAM.SFO");
	}
	printf("icon0.pngł܂...\n");
    if (icon0){
        t = fopen("../ICON0.PNG", "rb");
        fread(buffer, 1, icon0_size, t);
        fwrite(buffer, 1, icon0_size, out);
        fclose(t);
    }
	printf("icon1.pmfł܂...\n");
	if (icon1){
        t = fopen("../ICON1.PMF", "rb");
        fread(buffer, 1, icon1_size, t);
        fwrite(buffer, 1, icon1_size, out);
        fclose(t);
    }
    printf("pic0.pngł܂...\n");
 
	if (pic0){
        t = fopen("../PIC0.PNG", "rb");
        fread(buffer, 1, pic0_size, t);
        fwrite(buffer, 1, pic0_size, out);
        fclose(t);
    }
 
    printf("pic1.pngł܂...\n");
	if (pic1){
        t = fopen("../PIC1.PNG", "rb");
        fread(buffer, 1, pic1_size, t);
        fwrite(buffer, 1, pic1_size, out);
        fclose(t);
    }
        
	printf("snd0.at3ł܂...\n");
	if (snd){
        t = fopen("../SND0.AT3", "rb");
        fread(buffer, 1, snd_size, t);
        fwrite(buffer, 1, snd_size, out);
        fclose(t);
    }
	
    printf("data.pspł܂...\n");
	if (prx){
        t = fopen("DATA.PSP", "rb");
        fread(buffer, 1, prx_size, t);
        fwrite(buffer, 1, prx_size, out);
        fclose(t);
    }

	offset = ftell(out);

	for (i = 0; i < header[9]-offset; i++)
	{
		fputc(0, out);
	}
	

	fwrite("PSISOIMG0000", 1, 12, out);

	p1_offset = ftell(out);

	x = isosize + 0x100000;
	fwrite(&x, 1, 4, out);

	x = 0;
	for (i = 0; i < 0xFC; i++)
	{
		fwrite(&x, 1, 4, out);
	}

	memcpy(data1+1, code, 4);
	memcpy(data1+6, code+4, 5);

	//if (toc)
	if (toc == 2){
		memcpy(data1+1024, tocptr, toc_size);
		free(tocptr);
	}

	fwrite(data1, 1, sizeof(data1), out);

	p2_offset = ftell(out);
	x = isosize + 0x100000 + 0x2d31;
	fwrite(&x, 1, 4, out);

	strcpy((char *)data2+8, title);
	fwrite(data2, 1, sizeof(data2), out);

	index_offset = ftell(out);
	

	memset(dummy, 0, sizeof(dummy));

	offset = 0;

	if (complevel == 0)
	{
		x = 0x9300;
	}
	else
	{
		x = 0;
	}

	for (i = 0; i < isosize / 0x9300; i++)
	{
		fwrite(&offset, 1, 4, out);
		fwrite(&x, 1, 4, out);
		fwrite(dummy, 1, sizeof(dummy), out);

		if (complevel == 0)
			offset += 0x9300;
	}

	offset = ftell(out);

	for (i = 0; i < (header[9]+0x100000)-offset; i++)
	{
		fputc(0, out);
	}
	
	printf("C[Wt@Cł܂...\n");
	printf("  ݒ... [%04dMB|0000MB]",(isorealsize/(1024*1024)));

    int written=0;

	if (complevel == 0)
	{
		while ((x = fread(buffer, 1, 1048576, in)) > 0)
		{
			fwrite(buffer, 1, x, out);
         written+=x;
         printf("\010\010\010\010\010\010\010");
         printf("%04dMB]",((isorealsize-written)/(1024*1024)));
		}

		for (i = 0; i < (isosize-isorealsize); i++)
		{
			fputc(0, out);
		}

      printf("\n");
	}
	else
	{
		indexes = (IsoIndex *)malloc(sizeof(IsoIndex) * (isosize/0x9300));

		if (!indexes)
		{
			fclose(in);
			fclose(out);

			//ErrorExit("Failed in allocating Memory for Indexes!\n");
		}

		i = 0;
      int read=0;
		offset = 0;

		while ((x = fread(buffer2, 1, 0x9300, in)) > 0)
		{
			if (x < 0x9300)
			{
				memset(buffer2+x, 0, 0x9300-x);
			}

         read=x;

			x = deflateCompress(buffer2, 0x9300, buffer, sizeof(buffer), complevel);

			if (x < 0)
			{
				fclose(in);
				fclose(out);
				free(indexes);

				//ErrorExit("Error in Compression!\n");
			}

			memset(&indexes[i], 0, sizeof(IsoIndex));

			indexes[i].offset = offset;

			if (x >= 0x9300) /* Block didn't compress */
			{
				indexes[i].length = 0x9300;
				fwrite(buffer2, 1, 0x9300, out);
				offset += 0x9300;
			}
			else
			{
				indexes[i].length = x;
				fwrite(buffer, 1, x, out);
				offset += x;
			}

         written+=read;
         printf("\010\010\010\010\010\010\010");
         printf("%04dMB]",((isosize-written)/(1024*1024)));
			i++;
		}

      printf("\010\010\010\010\010\010\010");
      printf("  ]\n");

		if (i != (isosize/0x9300))
		{
			fclose(in);
			fclose(out);
			free(indexes);

			//ErrorExit("Some Error happened...\n");
		}

		x = ftell(out);

		if ((x % 0x10) != 0)
		{
			end_offset = x + (0x10 - (x % 0x10));

			for (i = 0; i < (end_offset-x); i++)
			{
				fputc('0', out);
			}
		}
		else
		{
			end_offset = x;
		}

		end_offset -= header[9];
	}
	
	printf("splash.pngł܂...\n");
 
 	fwrite("STARTDAT", 1, 8, out);
	char st_header[8]={0x01,0x00,0x00,0x00,0x01,0x00,0x00,0x00};
	
	fwrite(st_header,1,8,out);
	char splash_ad[4]={0x50,0x00,0x00,0x00};
	fwrite(splash_ad,1,4,out);
	
	t = fopen("../SPLASH.PNG", "rb");
    if (t){
            splash_size = getsize(t);
            splash = 1;
            fclose(t);
    }else{
            splash_size = 0;
    }
	
	fwrite(&splash_size,1,4,out);
	for (i = 0; i < 0x50-24; i++){
         fputc(0, out);
    }
	
	if (splash){
        t = fopen("../SPLASH.PNG", "rb");
        fread(buffer, 1, splash_size, t);
        fwrite(buffer, 1, splash_size, out);
        fclose(t);
    }
	
	if (complevel != 0){
		printf("kł܂...\n");

		fseek(out, p1_offset, SEEK_SET);
		fwrite(&end_offset, 1, 4, out);

		end_offset += 0x2d31;
		fseek(out, p2_offset, SEEK_SET);
		fwrite(&end_offset, 1, 4, out);

		fseek(out, index_offset, SEEK_SET);
		fwrite(indexes, 1, sizeof(IsoIndex) * (isosize/0x9300), out);
	}

	fclose(in);
	fclose(out);
}
int main(){
	printf("PS3-Free popstation v5 by tsun\n");
	int i=0;
	char iso_name[256]={0};
	char output[256]={0};
	char title[0x80]={0};
	char utf_title[0x80]={0};
	char save[0x80]={0};
	char utf_save[0x80]={0};
	char parental[8]={0};
	char region[8]={0};
	int complevel=0;
	char code[10];
	
	
	FILE *fp=fopen("__temp.dat","rb");
	if(fp){
		char buf[1];
		int k=0;
		
		//t@C
		fread(buf,1,1,fp);
		while(buf[0]!=0x2C){
			iso_name[i]=buf[0];
			fread(buf,1,1,fp);
			i++;
		}
		iso_name[i]=0x00;
		i=0;
		printf("t@CF%s\n",iso_name);
		//o͖
		fread(buf,1,1,fp);
		while(buf[0]!=0x2C){
				output[i]=buf[0];
			fread(buf,1,1,fp);
			i++;
		}
		output[i]=0x00;
		i=0;
		printf("t@CF%s\n",iso_name);
		//^Cg(SJIS)\p
		fread(buf,1,1,fp);
		while(buf[0]!=0x2C){
			title[i]=buf[0];
			i++;
			fread(buf,1,1,fp);
		}
		title[i]=0x00;
		printf("^CgF%s\n",title);
		i=0;
		//^Cg(UTF-8)
		fread(buf,1,1,fp);
		while(buf[0]!=0x2C){
			utf_title[i]=buf[0];
			i++;
			fread(buf,1,1,fp);
		}
		utf_title[i]=0x00;
		for(i=0;i<=strlen(utf_title);i++){
			pm_s[5][i]=utf_title[i];
		}
		i=0;
		//Z[u^Cg(SJIS)\p
		fread(buf,1,1,fp);
		while(buf[0]!=0x2C){
			save[i]=buf[0];
			i++;
			fread(buf,1,1,fp);
		}
		save[i]=0x00;
		printf("Z[u^CgF%s\n",save);
		i=0;
		//Z[u^Cg(UTF-8)
		fread(buf,1,1,fp);
		while(buf[0]!=0x2C){
			utf_save[i]=buf[0];
			i++;
			fread(buf,1,1,fp);
		}
		utf_save[i]=0x00;
		i=0;

		//ID		
		fread(buf,1,1,fp);
		while(buf[0]!=0x2C){
			code[i]=buf[0];
			i++;
			fread(buf,1,1,fp);
		}
		code[9]=0;
		for(i=0;i<9;i++){
			pm_s[2][i]=code[i];
		}
		pm_s[2][9]=0;
		printf("DISC IDF%s\n",code);
		i=0;
		
		//x
		fread(buf,1,1,fp);
		while(buf[0]!=0x2C){
			parental[i]=buf[0];
			i++;
			fread(buf,1,1,fp);
		}
		parental[i]=0x00;
		for(k=0;k<i;k++){
			int b=1;
			int j=0;
			for(j=0;j<i-k-1;j++)b*=10;
			pm_b[3]+=(parental[k]-'0')*b;
		}
		i=0;
		printf("xF%s\n",parental);
		//[WR[h
		fread(buf,1,1,fp);
		while(buf[0]!=0x2C){
			region[i]=buf[0];
			i++;
			fread(buf,1,1,fp);
		}
		region[i]=0x00;
		for(k=0;k<i;k++){
			int b=1;
			int j=0;
			for(j=0;j<i-k-1;j++)b*=16;
			pm_b[4]+=(region[k]-'0')*b;
		}
		i=0;
		printf("[WR[hF%s\n",region);
		//kx
		fread(buf,1,1,fp);
		complevel=(buf[0]-'0');
		printf("kxF%d\n",complevel);
		
		fclose(fp);
		
		remove("__temp.dat");

		printf("\nϊJn܂B\n");
	    convert(iso_name, output,utf_save,code,complevel);

	    printf("܂B\n");
	    return 0;
		
	}else{
		return 0;
	}
}

